"""VirtualHostFolder module

Defines the VirtualHostFolder class and the VirtualHost class

Based, obviously, heavily on VirtualHostMonster--thanks to Evan Simpson and
iMeme.  Comments and suggestions to garyposter@earthlink.net  Thanks for trying
it out and enjoy.

Requires Python 2.x
"""



# Why are we not just using OrderedFolder with all_meta_types controlled?
# Because we don't want VirtualHost objects to be a true product, to be added
# anywhere.  I haven't seen another solution to this, and the Developer's Guide
# seems to suggest that there is not another one.  In any case, here is what I
# decided to do.

from Globals import DTMLFile, MessageDialog, Persistent, PersistentMapping, ImageFile, InitializeClass
from OFS.SimpleItem import SimpleItem # Item
from Acquisition import Implicit, ImplicitAcquisitionWrapper
from ExtensionClass import Base
from ZPublisher import BeforeTraverse
from utils import VirtualHostKeyEmptyMarker
from App.Management import Navigation, Tabs
from AccessControl import ClassSecurityInfo
import re, OFS.PropertyManager # took out os

# from AccessRule import _swallow

def getErrorText(e):
    if type(e)==type(""): return [e]
    res=[]
    if hasattr(e,"args") and e.args:
        if e.args==type(""): res.append(e.args)
        else: res.extend(e.args)
    if hasattr(e,"text") and e.text:
        if e.text==type(""): res.append(e.text)
        else: res.extend(e.text)
    return res

_marker=[]

MANAGE_SECURITY_NAME="Manage Virtual Hosts"
TEST_SECURITY_NAME="Test Virtual Host Folder"

class VirtualHost(SimpleItem, Navigation, Tabs, OFS.PropertyManager.PropertyManager):
    """Represents one virtual host; can have multiple domains, but only one root
    """


    security = ClassSecurityInfo()
    security.declareObjectPublic()

    meta_type="Virtual Host"

    manage_options=(
        (
        {'label':'Configuration', 'action':'manage_main',
        'help':('VirtualHostFolder','VHF_Configure.stx')},
        )
        +OFS.PropertyManager.PropertyManager.manage_options
        +SimpleItem.manage_options
        )

    _properties=({'id':'title', 'type': 'string', 'mode': 'w'},)

    isUsefulTemplate=re.compile(r'\\(\d|g)') # not familiar with \g but Beazley lists it as an option (Python Essential Reference 2nd Ed p.140)
    isIPAddress=re.compile(r'^[\d.]+$')
    cleanPath=re.compile(r'^(.+?)/?$')

    # work with the domain list is always case insensitive
    # work with the domainRE is case insensitive by default but modifiable

    ## and now we need UI methods

    security.declareProtected( 'Manage Virtual Hosts', 'manage_main' )
    manage_main = DTMLFile('dtml/HostManage_main', globals())

    security.declareProtected( 'Manage Virtual Hosts', 'manage_host' )
    def manage_host(self, domainList=[], virtualRoot="", domainRE="", domainREIgnoreCase=0, virtualRootTemplate="", REQUEST=None):
        """the main management function.  The mfunc uses the optional
        messaging system: if a callable object is passed to the various
        worker-bee functions, they will call the function with messaging
        information.  In this case, we are having the worker-bee functions
        append to a messages array that we then return to the caller."""
        messages=[]
        mfunc=messages.append
        self.delDomainsFromList([x for x in self.getDomainList() if x not in domainList], mfunc)
        self.addDomainsToList([x for x in domainList if x not in self.getDomainList()], mfunc)
        if virtualRoot != self.getVirtualRoot():
            self.setVirtualRoot(virtualRoot, mfunc)
        if domainRE!=self.getDomainREComponents()[0] or ((domainREIgnoreCase and 1 or 0) ^ (self.getDomainREComponents()[1] and 1 or 0)):
            self.setDomainRE(domainRE, ignorecase=domainREIgnoreCase, mfunc=mfunc)
        if virtualRootTemplate != self.getVirtualRootTemplate():
            self.setVirtualRootTemplate(virtualRootTemplate, mfunc)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, messages=messages)
        return messages

    ## the non-UI basic methods

    def __init__(self, id, title=""):
        """yup, it's an init."""
        self.id=id
        self.title=title
        self._virtualRoot="/"
        self._domainList=()
        self._virtualRootTemplate=""
        self._domainRE=None

        ## the setters and getters.  I decided to not be so demanding on
        ## Zope and uncomment the persistence mechanism markers on the setters

    security.declarePrivate( 'setVirtualRoot' )
    def setVirtualRoot(self, virtualRoot="/", mfunc=None):
        """sets the static virtual root.  Uses optional message system (mfunc)."""
        if not virtualRoot: virtualRoot="/"
        virtualRoot=self.cleanPath.match(virtualRoot).group(1) # eliminates final slash if present
        self._virtualRoot = self._check_path(virtualRoot, mfunc) and virtualRoot or self._virtualRoot
        self._p_changed = 1
        if mfunc and self._virtualRoot==virtualRoot:
            if self._virtualRoot=="/":
                mfunc("Virtual host root set to Zope root")
            else:
                mfunc("Virtual host root changed")
        return self._virtualRoot==virtualRoot

    security.declarePublic( 'getVirtualRoot' ) # magic
    def getVirtualRoot(self, matchObj=None, matchObj2=None):
        """a magic function.  Call it with no arguments, and you will get the
        static path.  Call it with a regular expression match object, and it
        will try to return the expanded (filled-in) version of the dynamic
        virtual root template; if the template does not exist or the expanded
        path does not resolve to a valid Zope folder, the static path is
        returned."""
        if matchObj and self._virtualRootTemplate:
            dynamicRoot=matchObj.expand(self._virtualRootTemplate)
            if self._check_path(dynamicRoot): return dynamicRoot
            if matchObj2: # experimental lowercase
                dynamicRoot=matchObj2.expand(self._virtualRootTemplate)
                if self._check_path(dynamicRoot): return dynamicRoot
        return self._virtualRoot

    security.declarePrivate( 'addDomainsToList' )
    def addDomainsToList(self, domains, mfunc=None):
        """adds domains to list; uses optional message system"""
        domains=self._clean_domainList(domains) # shape into a lowercase tuple, remove blanks
        ln=len(self._domainList)
        self._domainList+=tuple([x for x in domains if x not in self._domainList and self._check_domain(x, mfunc)])
        ln=len(self._domainList)-ln
        if ln:
            self._p_changed = 1
            if mfunc:mfunc("%d domain%s added to domain list" % (ln, ln!=1 and "s" or ""))
        return ln

    security.declarePrivate( 'delDomainsFromList' )
    def delDomainsFromList(self, domains, mfunc=None):
        """deletes domains from list; uses optional message system"""
        domains=self._clean_domainList(domains) # shape into a lowercase tuple
        ln=len(self._domainList)
        self._domainList=tuple([x for x in self._domainList if x not in domains])
        ln=ln-len(self._domainList)
        if ln:
            self._p_changed = 1
            if mfunc:mfunc("%d domain%s deleted from domain list" % (ln, ln!=1 and "s" or ""))
        return ln

    security.declarePublic( 'getDomainList' )
    def getDomainList(self):
        """returns an unsorted domain list"""
        return self._domainList

    security.declarePublic( 'getSortedDomainList' )
    def getSortedDomainList(self):
        """returns a sorted version of the domain list"""
        x=list(self._domainList)
        x.sort()
        return x

    security.declarePrivate( 'setVirtualRootTemplate' )
    def setVirtualRootTemplate(self, tmpl, mfunc=None):
        """sets the virtual root template; supports the optional
        messaging system."""
        if not tmpl:
            if mfunc:mfunc("Virtual root template deleted.")
            self._virtualRootTemplate=""
            self._p_changed = 1
            return 1
        virtualRoot=self.cleanPath.match(tmpl).group(1) # eliminates final slash if present
        if self.isUsefulTemplate.search(tmpl):
            self._virtualRootTemplate=tmpl
            if mfunc:mfunc("Virtual root template changed")
            self._p_changed = 1
            return 1
        if mfunc:mfunc(r"Virtual rule template not set: the template must use back references such as \1 or else it is not a dynamic template.  Just set the standard virtual root instead.  If you want the regular expression to go to a different standard virtual root, consider creating a new host.")
        return 0

    security.declarePublic( 'getVirtualRootTemplate' )
    def getVirtualRootTemplate(self):
        """returns the unprocessed vitual root template"""
        return self._virtualRootTemplate

    security.declarePrivate( 'setDomainRE' )
    def setDomainRE(self, domainRE, ignorecase=1, mfunc=None):
        """sets the domain regular expression.  Pass in anything that evaluates
        to Boolean true in ignorecase to set it to true.  No other flags are
        available (although they could be added, I didn't feel they would be
        useful).  The messages array is an optional human-readable
        messaging system described in the doc string for _check_path below: feel
        free to ignore."""
        if domainRE:
            test=0
            try:
                test=re.compile(domainRE, ignorecase and re.I or 0)
            except (SyntaxError, re.error), e:
                # get_transaction().abort() # Chris Withers Nov 25 2001 Zope digest
                if mfunc:mfunc("The domain regular expression could not be changed. The most likely error is a syntax error in your regular expression.  Python reports the following: %s" % "; ".join(getErrorText(e)))
                return 0
            if mfunc:mfunc("The domain regular expression is %s" % (self._domainRE and "changed" or "added"))
            self._domainRE=test
        else:
            self._domainRE=None
            if mfunc:mfunc("The domain regular expression is deleted")
        self._p_changed = 1
        return 1

    security.declarePublic( 'getDomainRE' )
    def getDomainRE(self):
        """returns the current domain regular expression *object*"""
        return self._domainRE

    security.declarePublic( 'getDomainREComponents' )
    def getDomainREComponents(self):
        """returns the current domain regular expression and the ignorecase
        flag setting.  If no current domain regular expression is set, returns an empty string and the default 'on' ignorecase setting."""
        return self._domainRE and (self._domainRE.pattern, self._domainRE.flags and 1) or ("", 1)

    def _matchDomainRE(self, domain):
        """silently handles absence of RE; included in this group because it
        directly accesses the _domainRE rather than using the getter.  Perhaps
        a weird, misguided combination of anal and faux-speed conscious, at
        worst."""
        return self._domainRE and self._domainRE.match(domain) or None

        ## the other basic method(s)
    security.declarePublic( 'getVirtualRootByDomain' )
    def getVirtualRootByDomain(self, domain):
        """This is the heart of the host object: the method by which you can
        pass in a domain and get either a path for that domain (also signifying
        that this is the pertinent host object for the domain) or None."""
        domain=domain.strip()
        lowdom=domain.lower()
        if lowdom in self.getDomainList(): return self.getVirtualRoot()
        res = self._matchDomainRE(domain)
        if res:
            res2= (domain != lowdom) and self._matchDomainRE(lowdom)
            return self.getVirtualRoot(res, res2)
        return None

        ## the utility methods (should only be called internally to object)
    def _check_path(self,path,mfunc=None):
        """pass in a path, have it validated as bad (empty, not found, or not
        a folder) (0) or ok (1).  The optional mfunc thing is a trick to let the method return a clear on-off numeric success message, while allowing it also to communicate some human-readable messages should we want them.  Just assign a variable in your local scope to an array, pass it along with the path, and then the next time you look at it, through the magic of passing by reference, it might have a message from this function.  This particular function will currently only give you a message if it encounters a bad path."""
        if not path:
           if mfunc:mfunc("Empty path ignored" ) # I think I heard append is costly; I wonder if I should set messages to None for all of these and then do a if messages: messages.append thing?
           return 0
        if path[-1] == '/':
           if len(path) == 1:
              return 1
           else:
              path = path[:-1]
        try:
            leaf_object = self.unrestrictedTraverse(path)
        except:
            if mfunc:mfunc("The path %s was not found, so ignored" % path)
            return 0
        if not getattr(leaf_object.aq_base,'isAnObjectManager', None):
            if mfunc:mfunc( "Paths must end in a container: %s does not." % path )
        return 1

    def _check_domain(self, domain, mfunc=None):
        """This is a plug to put in code that rejects domains.  At the
        moment, all it does is try to make sure the domain is not a numeric IP
        address.  That's probably all it will ever need to do too, but hey, here
        it is."""
        if self.isIPAddress.match(domain):
            if mfunc:mfunc("%s : IP addresses are not mappable" % domain)
            return 0
        return 1

    def _clean_domainList(self, domains, mfunc=None):
        """a utility function.  converts all of the domain names to lower
        case, which is convenient for the domain list functionality, and does
        whatever else needs doing to the list to shape it up."""
        return [x.lower().strip() for x in domains if x and not x.isspace() ]

    def getId(self): # why did I do this?
        return self.id


class VirtualHostFolder(SimpleItem, Navigation, Tabs): # should be ordered...shouldn't be a true object manager...
    """Provide a drop-in solution for virtual hosting.
    """


    security = ClassSecurityInfo()
    security.declareObjectPublic()

    meta_type='Virtual Host Folder'
    id='VHF' # the thought was to force this...not doing that right now...
    priority = 25
    isPrincipiaFolderish=1
    title = 'Virtual Host Folder'

    def __init__(self):
        self._hosts=()

    manage_options=(
        (
        {'label':'Contents', 'action':'manage_main', 'help':('VirtualHostFolder','VHF_Edit.stx')},
        {'label':'Test', 'action':'manage_testDomainForm',
        'help':('VirtualHostFolder','VHF_TestDomainPath.stx')}
        )
        +SimpleItem.manage_options
        ) # removed "+RoleManager.manage_options"

    security.declareProtected( 'Manage Virtual Hosts', 'manage_main' )
    manage_main = DTMLFile('dtml/VHFmanage_main', globals())

    security.declareProtected( 'Test Virtual Host Folder', 'manage_testDomainForm' )
    manage_testDomainForm = DTMLFile('dtml/VHFtestDomain', globals())

    security.declareProtected( 'Manage Virtual Hosts', 'manage_addHostForm' )
    manage_addHostForm = DTMLFile('dtml/addVirtualHost', globals())

    security.declareProtected('Manage Virtual Hosts', 'manage_renameHostsForm' )
    manage_renameHostsForm=DTMLFile('dtml/renameVirtualHostsForm', globals())

    # created to try and track down unrestrictedTraverse problem in path template check
    #security.declareProtected( 'Test Virtual Host Folder', 'manage_testDomainForm' )
    #def manage_testDomain(self, test_domain="", REQUEST=None):
        #"""handle the actual testing of a domain from the management interface
        #manage_testDomainForm"""
        #hostObj, rootPath=self.getHostAndRootByDomain(test_domain)
        #if hostObj:
            #message="%s resolves to the root path %s, as defined by the host %s." % (test_domain, rootPath, hostObj.id)
        #else:
            #message="%s does not match any domain in the virtual host folder.  The folder would send the user to the Zope root." % test_domain
        #if REQUEST is not None:
            #return MessageDialog(title='Path Test Results',
              #message=message,
              #action='./manage_testDomainForm')

    security.declareProtected( 'Manage Virtual Hosts', 'manage_renameHosts' )
    def manage_renameHosts(self, ids=[], new_ids=[], REQUEST=None):
        """Rename several hosts"""
        if len(ids) != len(new_ids):
            raise 'Bad Request','Please rename each listed virtual host; deleting must be done separately.'
        result=self.renameHostIds([(ids[x], new_ids[x]) for x in range(len(ids))], REQUEST=REQUEST)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, update_menu=1, message='%d hosts renamed' % result)
        return '%d hosts renamed' % result

    security.declarePublic( 'manage_getObjsByIds' )
    def manage_getObjsByIds(self, REQUEST):
        """returns objects whose ids are listed in the REQUEST"""
        ids=REQUEST["ids"]
        return [x for x in self.getHosts() if x.id in ids]

    security.declareProtected( 'Manage Virtual Hosts', 'manage_delHosts' )
    def manage_delHosts(self, ids=[], REQUEST=None):
        """deletes selected hosts; returns HTML or text confirmation"""
        if type(ids) is type(''): ids=[ids]
        if not ids:
            if REQUEST is not None:
                return self.manage_main(self, REQUEST, message='No items were specified!')
            else: return MessageDialog(title='No items specified',
                   message='No items were specified!',
                   action ='./manage_main',)
        attempt=len(ids)
        final=self.delHostsByIds(ids)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, update_menu=1, message="Performed %d of %d requested deletions." % (final, attempt)) # update_menu? The pertinent code isn't in my dtml page but I'll leave it in case I want to add it in for some reason.
        return "Performed %d of %d requested deletions." % (final, attempt)

    security.declareProtected( 'Manage Virtual Hosts', 'manage_moveHostsUp' )
    def manage_moveHostsUp(self, ids=[], REQUEST=None):
        """moves selected hosts up by one, if possible, in the host
        folder; returns HTML or text confirmation"""
        if type(ids) is type(''): ids=[ids]
        if not ids:
            if REQUEST is not None:
                return self.manage_main(self, REQUEST, message='No items were specified!')
            else: return MessageDialog(title='No items specified',
                   message='No items were specified!',
                   action ='./manage_main',)
        attempt=len(ids)
        self.moveHostsUpByIds(ids)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, message="%d items moved up." % attempt)
        return "%d items moved up." % attempt

    security.declareProtected( 'Manage Virtual Hosts', 'manage_moveHostsDown' )
    def manage_moveHostsDown(self, ids=[], REQUEST=None):
        """moves selected hosts down by one, if possible, in the host
        folder; returns HTML or text confirmation"""
        if type(ids) is type(''): ids=[ids]
        if not ids:
            if REQUEST is not None:
                return self.manage_main(self, REQUEST, message='No items were specified!')
            else: return MessageDialog(title='No items specified',
                   message='No items were specified!',
                   action ='./manage_main',)
        attempt=len(ids)
        self.moveHostsDownByIds(ids)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, message="%d items moved down." % attempt)
        return "%d items moved down." % attempt

    security.declareProtected( 'Manage Virtual Hosts', 'manage_moveHostsToTop' )
    def manage_moveHostsToTop(self, ids=[], REQUEST=None):
        """moves selected hosts to top of host folder; returns HTML or
        text confirmation"""
        if type(ids) is type(''): ids=[ids]
        if not ids:
            if REQUEST is not None:
                return self.manage_main(self, REQUEST, message='No items were specified!')
            else: return MessageDialog(title='No items specified',
                   message='No items were specified!',
                   action ='./manage_main',)
        attempt=self.moveHostsToTopByIds(ids)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, message="%d items moved to top." % attempt)
        return "%d items moved to top." % attempt

    security.declareProtected( 'Manage Virtual Hosts', 'manage_moveHostsToBottom' )
    def manage_moveHostsToBottom(self, ids=[], REQUEST=None):
        """moves selected hosts to bottom of host folder; returns HTML or
        text confirmation"""
        if type(ids) is type(''): ids=[ids]
        if not ids:
            if REQUEST is not None:
                return self.manage_main(self, REQUEST, message='No items were specified!')
            else: return MessageDialog(title='No items specified',
                   message='No items were specified!',
                   action ='./manage_main',)
        attempt=self.moveHostsToBottomByIds(ids)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, message="%d items moved to bottom." % attempt)
        return "%d items moved to bottom." % attempt

    security.declareProtected( 'Manage Virtual Hosts', 'manage_addHost' )
    def manage_addHost(self, id, title="", REQUEST=None):
        """adds a host to the host folder; returns HTML or text confirmation"""
        res=self.addHost(id, title)
        if not res:
            if REQUEST is not None:
                return self.manage_main(self, REQUEST, message='No items were specified!')
            else: return MessageDialog(title='Error',
                   message='ID is in use or invalid; please try another',
                   action ='./manage_addHost',)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, message="Host %s added at bottom of list." % id)
        return "Host %s added at bottom of list." % id

    security.declarePublic( 'getHosts' )
    def getHosts(self):
        """returns tuple of hosts in user-blessed order"""
        return self._hosts

    security.declarePublic( 'getAcquiredHosts' )
    def getAcquiredHosts(self):
        """returns list of hosts in user-blessed order with acquisition set"""
        return [x.__of__(self) for x in self._hosts]

    security.declarePublic( 'getHostKeyValTuples' )
    def getHostKeyValTuples(self):
        """returns a list of tuples, in which each tuple is the id and actual
        host object of one of the hosts in the folder"""
        return [(x.id, x) for x in self.getHosts()]

    security.declarePublic( 'getAcquiredHostKeyValTuples' )
    def getAcquiredHostKeyValTuples(self):
        """returns a list of tuples, in which each tuple is the id and actual
        *acquired* host object of one of the hosts in the folder"""
        return [(x.id, x.__of__(self)) for x in self._hosts]

    security.declarePublic( 'getHostAndRootByDomain' )
    def getHostAndRootByDomain(self,domain):
        """pass in a domain, get a tuple of the acquired host object and the
        root path that are associated with the domain, if any (otherwise you get
        a tuple of two Nones."""
        for host in self.getAcquiredHosts():
            root=host.getVirtualRootByDomain(domain)
            if root: return host, root
        return None, None

    security.declarePublic( 'getHostById' )
    def getHostById(self,id):
        """pass in a string id, get the associated host object or None if not
        found"""
        for host in self.getHosts():
            if host.id==id: return host.__of__(self)
        return None

    security.declarePublic( 'getCurrentHost' )
    def getCurrentHost(self, REQUEST=None):
        """Pass in the request object (and if you don't, it'll still try) and
        get back the currently active host object, if any, or None"""
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        if REQUEST and REQUEST.has_key("VIRTUAL_HOST_ID"):
            return self.getHostById(REQUEST["VIRTUAL_HOST_ID"])
        return None

    def _util_getMoveHostsUp(self, hosts, hostIds): # hosts as tuple of VirtualHosts, hostIds as tuple of id strings
        """a utility function used by the host ordering functions.  Don't call
        this ever: this is private.  Use the host ordering functions instead.
        Just so you know, you pass in a tuple or list of host objects in hosts
        and a tuple or list of string ids in hostIds.  It returns a list of
        hosts (yes, a list) in which the hosts with ids matching those
        in the hostIds list have been moved up by one, if possible.  This is used by the MoveUp and MoveDown ordering functions."""
        res=[]
        last=None
        for host in hosts:
            if host.id in hostIds:
                res.append(host)
            else:
                if last: res.append(last)
                last=host
        res.append(last)
        return res

    def _util_separateHosts(self,hosts,hostIds):
        """a utility function: private, hands off.  Pass in a list or tuple of
        host objects in hosts and a list or tuple of string ids in hostIds, get
        back a tuple of two lists of host objects: the first list contains those
        hosts that have ids matching one in the hostIds list, the second holds
        the remainder.  The function retains the ordering of hosts within the
        two groups.  It is used by the ToTop and ToBottom functions."""
        mem=[]
        out=[]
        for host in hosts:
            if host.id in hostIds: mem.append(host)
            else: out.append(host)
        return mem, out

    security.declarePrivate( 'moveHostsUpByIds' )
    def moveHostsUpByIds(self, hostIds):
        """the nuts-and-bolts version of moving the hosts up"""
        self._hosts=tuple(self._util_getMoveHostsUp(self.getHosts(),hostIds))
        self._p_changed = 1
        return 1

    security.declarePrivate( 'moveHostsDownByIds' )
    def moveHostsDownByIds(self, hostIds):
        """the nuts-and-bolts version of moving the hosts down"""
        x=list(self.getHosts())
        x.reverse()
        x=self._util_getMoveHostsUp(x,hostIds)
        x.reverse()
        self._hosts=tuple(x)
        self._p_changed = 1
        return 1

    security.declarePrivate( 'moveHostsToTopByIds' )
    def moveHostsToTopByIds(self, hostIds):
        """the nuts-and-bolts version of moving the hosts to the top"""
        mem, out=self._util_separateHosts(self.getHosts(),hostIds)
        self._hosts=tuple(mem+out)
        self._p_changed = 1
        return len(mem)

    security.declarePrivate( 'moveHostsToBottomByIds' )
    def moveHostsToBottomByIds(self, hostIds):
        """the nuts-and-bolts version of moving the hosts to the bottom"""
        mem, out=self._util_separateHosts(self.getHosts(),hostIds)
        self._hosts=tuple(out+mem)
        self._p_changed = 1
        return len(mem)

    security.declarePrivate( 'delHostsByIds' )
    def delHostsByIds(self, hostIds):
        """the nuts-and-bolts version of deleting hosts by a list or tuple of
        ids"""
        l=len(self._hosts)
        self._hosts=tuple([x for x in self._hosts if x.id not in hostIds])
        self._p_changed = 1
        return l-len(self._hosts)

    security.declarePrivate( 'addHost' )
    def addHost(self, id, title=""):
        """the nuts-and-bolts version of adding a single host.  Always adds to
        the bottom of the list for now."""
        if not (hasattr(self, id) or self.getHostById(id) or id[:1]=="_"):
            self._hosts+=(VirtualHost(id, title),)
            self._p_changed = 1
            return 1
        return 0

    security.declarePrivate( 'renameHostIds' )
    def renameHostIds(self, oldNewIdPairs, REQUEST=None):
        """the nuts-and-bolts version of renaming hosts.  Tries to maintain the
        VIRTUAL_HOST_ID key as well, if you are manipulating the host used by
        the current domain.  Expects a sequence of sequence pairs: oldId, newId.  Example: (("oldID","newID"),("anotherOld","anotherNew"))"""
        count=0
        if not REQUEST and hasattr(self, "REQUEST"):
            REQUEST=self.REQUEST
        currentId=""
        if REQUEST and REQUEST.has_key("VIRTUAL_HOST_ID"):
            currentId=REQUEST["VIRTUAL_HOST_ID"]
        for oldId, id in oldNewIdPairs:
            if oldId != id and not (hasattr(self, id) or self.getHostById(id)):
                host=self.getHostById(oldId)
                if host:
                    host._setId(id)
                    count+=1
                    if REQUEST and currentId==oldId:
                        REQUEST.set( "VIRTUAL_HOST_ID", id )
        if count: self._p_changed = 1
        return count

    def __call__(self, client, request, response=None):
        '''Traversing at home'''
        stack = request['TraversalRequestNameStack']
        host = ''
        if stack and stack[-1] == 'VirtualHostBase':
            stack.pop()
            protocol = stack.pop()
            host = stack.pop()
            if ':' in host:
                host, port = host.split(':')
                request.setServerURL(protocol, host, port)
            else:
                request.setServerURL(protocol, host)
        if host == '':
           host = request['HTTP_HOST'].split(':')[0]
        # If a domain to path mapping applies, invoke it now.
        if not stack.count('VirtualHostRoot'):
            hostObj, rootPath=self.getHostAndRootByDomain(host)
            request.set( "VIRTUAL_HOST_ID", hostObj and hostObj.id or VirtualHostKeyEmptyMarker )
            if hostObj:
                if rootPath != "/":
                    list_path = rootPath.split('/')[1:]
                    stack.append('/') # the flag that bobo_traverse looks for
                    stack.append(self.id)
                    for i in range(len(list_path)):
                        stack.append(list_path.pop())
        else:
            # below untouched other than placing in else clause and switching to
            # string methods.
            # all this indentation should mean is that you cannot put
            # "VirtualHostRoot" in the VHM paths and have it work...
            # which seems logical enough.

            # Find and convert VirtualHostRoot directive
            # If it is followed by one or more path elements that each
            # start with '_vh_', use them to construct the path to the
            # virtual root.
            vh = -1
            for ii in range(len(stack)):
                if stack[ii] == 'VirtualHostRoot':
                    if vh >= 0:
                        pp = ['']
                        for jj in range(vh, ii):
                            pp.insert(1, stack[jj][4:])
                        stack[vh:ii + 1] = [pp.join('/'), self.id]
                    elif ii > 0 and stack[ii - 1][:1] == '/':
                        stack[ii] = self.id
                    else:
                        stack[ii] = self.id
                        stack.insert(ii, '/')
                    break
                elif vh < 0 and stack[ii][:4] == '_vh_':
                    vh = ii

    def __bobo_traverse__(self, request, name): # mostly untouched, except for adding the getHostByID thing...
        '''Traversing away'''
        # we are doing normal traversal if the name isn't (or doesn't begin
        # with) a forward slash.
        if name[:1] != '/':
            if hasattr(self, name): return getattr(self, name)
            x= self.getHostById(name) # ... right here
            if x: return x.__of__(self)
            return None
        # else we are dealing with the part of the path that __call__ set up (or
        # that the manager hardcoded in the external server or included as
        # testing directly in the URL, via old-style SiteAccess usage)
        parents = request.PARENTS
        parents.pop() # I don't belong there

        if len(name) > 1:
            request.setVirtualRoot(name[1:].split('/'))
        else:
            request.setVirtualRoot([])
        return parents.pop() # He'll get put back on



#####################################################################

    # what do I do with these for security permissions?  I probably should specify them as private.

    def addToContainer(self, container):
        """the magic begins in manage_afterAdd"""
        container._setObject(self.id, self)
        self.manage_afterAdd(self, container)

    def manage_addToContainer(self, container, nextURL=''):
        """The Zope management wrapper around addToContainer: the main point is
        to eventually get to the after_add hook"""
        self.addToContainer(container)
        if nextURL:
            return MessageDialog(title='Item Added',
              message='This object now has a %s' % self.meta_type,
              action=nextURL)

    def manage_beforeDelete(self, item, container):
        """takes care of unhooking the traversal call"""
        if item is self:
            BeforeTraverse.unregisterBeforeTraverse(container, self.meta_type)

    def manage_afterAdd(self, item, container):
        """This hooks up the __call__ function as a way to pre-process the
        traversal of the path so it can modify things"""
        if item is self:
            id = self.id
            if callable(id): id = id()

            # We want the original object, not stuff in between
            container = container.this()
            hook = BeforeTraverse.NameCaller(id)
            BeforeTraverse.registerBeforeTraverse(container, hook,
                                                  self.meta_type,
                                                  self.priority)
    def _setId(self, id):
        """necessary for id changes; I better watch out to use this when
        necessary"""
        # why does it never actually change the id?
        id = str(id)
        if id != self.id:
            BeforeTraverse.unregisterBeforeTraverse(container,
            self.meta_type)
            hook = BeforeTraverse.NameCaller(id)
            BeforeTraverse.registerBeforeTraverse(container, hook,
                                                  self.meta_type,
                                                  self.priority)






manage_addVirtualHostFolderForm=DTMLFile('dtml/addVirtualHostFolder', globals())

def manage_addVirtualHostFolder(self, id, REQUEST=None, title="", **ignored):
    """Constructor for the Virtual Host Folder"""
    rootObj=self.getPhysicalRoot()
    if self.this()==rootObj and not rootObj.objectItems( "Virtual Host Folder" ) and not rootObj.objectItems( "Virtual Host Monster" ):
        vhf = VirtualHostFolder()
        vhf.id = str(id)
        if title: vhf.title=title
        if REQUEST:
            return vhf.manage_addToContainer(self.this(),
                                        '%s/manage_main' % REQUEST['URL1'])
        else:
            vhf.addToContainer(self.this())
            return 1
    if REQUEST:
        message="""ERROR: A Virtual Host Folder can only be added in the Zope
                root folder, and only if a Virtual Host Folder or Virtual Host
                Monster does not exist there already."""
        if rootObj.objectItems( "Virtual Host Monster" ):
            message+="""<br /><br />If you decide to uninstall the Virtual Host
            Monster, beware: you need to disable your Apache server settings
            <em>first</em>, delete the VHM or move it out of your root folder,
            install the VHF, and put the Apache httpd.conf file change in as
            listed in the "Add Virtual Host Folder" help page (in the help
            system) as the last step."""
        return MessageDialog(title='ERROR: Item NOT Added',
                message=message,
                action='%s/manage_main' % REQUEST['URL1'])
    else:
        vhf.addToContainer(self.this())
        return 1

constructors = (
  ('manage_addVirtualHostFolderForm', DTMLFile('www/addVirtualHostFolder', globals())),
  ('manage_addVirtualHostFolder', manage_addVirtualHostFolder),
)


InitializeClass(VirtualHost)
InitializeClass(VirtualHostFolder)
